import json
import sys
import traceback

import pytest

import apiritif
from apiritif.samples import ApiritifSampleExtractor, Sample


def pytest_addoption(parser):
    group = parser.getgroup('apiritif', 'reporting per-testcase Apiritif traces')
    group.addoption('--apiritif-trace', help='target path to save trace JSON')
    group.addoption('--apiritif-trace-detail', type=int, default=1, help='detail level for Apiritif trace (1-3)')


def pytest_configure(config):
    if not config.option.apiritif_trace:
        return
    plugin = ApiritifPytestPlugin(config)
    config.pluginmanager.register(plugin, plugin.__class__.__name__)


def pytest_unconfigure(config):
    name = ApiritifPytestPlugin.__name__
    if config.pluginmanager.has_plugin(name):
        plugin = config.pluginmanager.get_plugin(name)
        config.pluginmanager.unregister(plugin)


class ApiritifPytestPlugin(object):
    def __init__(self, config=None) -> None:
        super().__init__()
        self._result_file = None
        self._detail_level = 1
        if config:
            self._result_file = config.option.apiritif_trace
            self._detail_level = config.option.apiritif_trace_detail
        self._trace_map = {}

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_setup(self, item):
        self._pop_events()  # clean it, just in case
        yield

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        report = yield
        if call.when == 'call':
            self._trace_map[item.nodeid] = self._get_subsamples(call, report.get_result(), item)

    def _get_subsamples(self, call, report, item):
        recording = self._pop_events()
        sample = Sample()
        sample.test_case = item.name
        if item.parent:
            sample.test_suite = item.parent.name
        sample.duration = report.duration
        sample.status = 'PASSED' if report.passed else 'FAILED'
        if call.excinfo:
            sample.error_msg = str(call.excinfo.value)
            sample.error_trace = traceback.format_tb(call.excinfo.tb)

        extr = ApiritifSampleExtractor()
        trace = extr.parse_recording(recording, sample)
        toplevel_sample = trace[0].to_dict()
        self._filter([toplevel_sample])
        return toplevel_sample

    @pytest.hookimpl(tryfirst=True)
    def pytest_sessionfinish(self, session):
        with open(self._result_file, 'w') as fp:
            json.dump(self._trace_map, fp, indent=True)

    def _pop_events(self):
        return apiritif.recorder.pop_events(from_ts=-1, to_ts=sys.maxsize)

    def _filter(self, items):
        for item in items:
            self._filter(item['subsamples'])
            if self._detail_level >= 4:
                if isinstance(item['extras'].get('requestBody'), bytes):
                    try:
                        # optimistic
                        item['extras']['requestBody'] = item['extras']['requestBody'].decode('utf-8')
                    except UnicodeError:
                        # worst case it is just byte sequence
                        item['extras']['requestBody'] = item['extras']['requestBody'].decode('latin-1')

            if self._detail_level <= 3:
                item['extras'].pop('requestCookiesRaw', None)
                item['extras'].pop('requestCookies', None)
                item['extras'].pop('requestBody', None)
                item['extras'].pop('responseBody', None)
                item['extras'].pop('requestHeaders', None)
                item['extras'].pop('responseHeaders', None)

            if self._detail_level <= 2:
                item.pop('extras', None)
                item.pop('subsamples', None)
                item.pop('assertions', None)



"""

Copyright 2017 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import copy
import unicodecsv as csv
import json
import logging
import multiprocessing
import os
import sys
import time
import traceback
from multiprocessing.pool import ThreadPool
from optparse import OptionParser
from threading import Thread

from nose.config import Config, all_config_files
from nose.core import TestProgram
from nose.loader import defaultTestLoader
from nose.plugins import Plugin
from nose.plugins.manager import DefaultPluginManager

import apiritif
import apiritif.thread as thread
import apiritif.store as store
from apiritif.action_plugins import ActionHandlerFactory, import_plugins
from apiritif.utils import NormalShutdown, log, get_trace, VERSION


# TODO how to implement hits/s control/shape?
# TODO: VU ID for script
# TODO: disable assertions for load mode


def spawn_worker(params):
    """
    This method has to be module level function

    :type params: Params
    """
    setup_logging(params)
    log.info("Adding worker: idx=%s\tconcurrency=%s\tresults=%s", params.worker_index, params.concurrency,
             params.report)
    worker = Worker(params)
    worker.start()
    worker.join()


class Params(object):
    def __init__(self):
        super(Params, self).__init__()
        self.worker_index = 0
        self.worker_count = 1
        self.thread_index = 0
        self.report = None

        self.delay = 0

        self.concurrency = 1
        self.iterations = 1
        self.ramp_up = 0
        self.steps = 0
        self.hold_for = 0

        self.verbose = False

        self.tests = None

    def __repr__(self):
        return "%s" % self.__dict__


class Supervisor(Thread):
    """
    apiritif-loadgen CLI utility
        overwatch workers, kill them when terminated
        probably reports through stdout log the names of report files
    :type params: Params
    """

    def __init__(self, params):
        super(Supervisor, self).__init__(target=self._start_workers)
        self.setDaemon(True)
        self.setName(self.__class__.__name__)

        self.params = params
        self.workers = None

    def _concurrency_slicer(self, ):
        total_concurrency = 0
        inc = self.params.concurrency / float(self.params.worker_count)
        assert inc >= 1
        for idx in range(0, self.params.worker_count):
            progress = (idx + 1) * inc

            conc = int(round(progress - total_concurrency))
            assert conc > 0

            log.debug("Idx: %s, concurrency: %s", idx, conc)

            params = copy.deepcopy(self.params)
            params.worker_index = idx
            params.thread_index = total_concurrency  # for subprocess it's index of its first thread
            params.concurrency = conc
            params.report = self.params.report % idx
            params.worker_count = self.params.worker_count

            total_concurrency += conc

            yield params

        assert total_concurrency == self.params.concurrency

    def _start_workers(self):
        log.info("Total workers: %s", self.params.worker_count)

        thread.set_total(self.params.concurrency)
        self.workers = multiprocessing.Pool(processes=self.params.worker_count)
        args = list(self._concurrency_slicer())

        try:
            self.workers.map(spawn_worker, args)
        finally:
            self.workers.close()
            self.workers.join()
        # TODO: watch the total test duration, if set, 'cause iteration might last very long


class Worker(ThreadPool):
    def __init__(self, params):
        """
        :type params: Params
        """
        super(Worker, self).__init__(params.concurrency)
        self.params = params
        if self.params.report.lower().endswith(".ldjson"):
            store.writer = LDJSONSampleWriter(self.params.report)
        else:
            store.writer = JTLSampleWriter(self.params.report)

    def start(self):
        import_plugins()
        params = list(self._get_thread_params())
        with store.writer:  # writer must be closed finally
            try:
                self.map(self.run_nose, params)
            finally:
                self.close()

    def close(self):
        log.info("Workers finished, awaiting result writer")
        while not store.writer.is_queue_empty() and store.writer.is_alive():
            time.sleep(0.1)
        log.info("Results written, shutting down")
        super(Worker, self).close()

    def run_nose(self, params):
        """
        :type params: Params
        """
        thread.set_index(params.thread_index)
        log.debug("[%s] Starting nose iterations: %s", params.worker_index, params)
        assert isinstance(params.tests, list)
        # argv.extend(['--with-apiritif', '--nocapture', '--exe', '--nologcapture'])

        end_time = self.params.ramp_up + self.params.hold_for
        end_time += time.time() if end_time else 0
        time.sleep(params.delay)

        plugin = ApiritifPlugin()
        store.writer.concurrency += 1

        config = Config(env=os.environ, files=all_config_files(), plugins=DefaultPluginManager())
        config.plugins.addPlugins(extraplugins=[plugin])
        config.testNames = params.tests
        config.verbosity = 3 if params.verbose else 0
        if params.verbose:
            config.stream = open(os.devnull, "w")  # FIXME: use "with", allow writing to file/log

        iteration = 0
        handlers = ActionHandlerFactory.create_all()
        log.debug(f'Action handlers created {handlers}')
        thread.put_into_thread_store(action_handlers=handlers)
        for handler in handlers:
            handler.startup()

        try:
            while True:
                if 'GRACEFUL' in os.environ and os.path.exists(os.environ['GRACEFUL']):
                    break
                log.debug("Starting iteration:: index=%d,start_time=%.3f", iteration, time.time())
                thread.set_iteration(iteration)
                ApiritifTestProgram(config=config)
                log.debug("Finishing iteration:: index=%d,end_time=%.3f", iteration, time.time())

                iteration += 1

                # reasons to stop
                if plugin.stop_reason:
                    log.debug("[%s] finished prematurely: %s", params.worker_index, plugin.stop_reason)
                elif 0 < params.iterations <= iteration:
                    log.debug("[%s] iteration limit reached: %s", params.worker_index, params.iterations)
                elif 0 < end_time <= time.time():
                    log.debug("[%s] duration limit reached: %s", params.worker_index, params.hold_for)
                else:
                    continue  # continue if no one is faced

                break
        finally:
            store.writer.concurrency -= 1

            if params.verbose:
                config.stream.close()

            for handler in handlers:
                handler.finalize()

    def __reduce__(self):
        raise NotImplementedError()

    def _get_thread_params(self):
        if not self.params.steps or self.params.steps < 0:
            self.params.steps = sys.maxsize

        step_granularity = self.params.ramp_up / self.params.steps
        ramp_up_per_thread = self.params.ramp_up / self.params.concurrency
        for thr_idx in range(self.params.concurrency):
            offset = self.params.worker_index * ramp_up_per_thread / float(self.params.worker_count)
            delay = offset + thr_idx * float(self.params.ramp_up) / self.params.concurrency
            delay -= delay % step_granularity if step_granularity else 0
            params = copy.deepcopy(self.params)
            params.thread_index = self.params.thread_index + thr_idx
            params.delay = delay
            yield params


class ApiritifTestProgram(TestProgram):
    def __init__(self, *args, **kwargs):
        super(ApiritifTestProgram, self).__init__(*args, **kwargs)
        self.testNames = None

    def parseArgs(self, argv):
        self.exit = False
        self.testNames = self.config.testNames
        self.testLoader = defaultTestLoader(config=self.config)
        self.createTests()


class LDJSONSampleWriter(object):
    """
    :type out_stream: file
    """

    def __init__(self, output_file):
        super(LDJSONSampleWriter, self).__init__()
        self.concurrency = 0
        self.output_file = output_file
        self.out_stream = None
        self._samples_queue = multiprocessing.Queue()

        self._writing = False
        self._writer_thread = Thread(target=self._writer)
        self._writer_thread.setDaemon(True)
        self._writer_thread.setName(self.__class__.__name__)

    def __enter__(self):
        self.out_stream = open(self.output_file, "wb")
        self._writing = True
        self._writer_thread.start()
        return self

    def is_alive(self):
        return self._writer_thread.is_alive()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._writing = False
        self._writer_thread.join()
        self.out_stream.close()

    def add(self, sample, test_count, success_count):
        self._samples_queue.put_nowait((sample, test_count, success_count))

    def is_queue_empty(self):
        return self._samples_queue.empty()

    def _writer(self):
        while self._writing:
            if self._samples_queue.empty():
                time.sleep(0.1)

            while not self._samples_queue.empty():
                item = self._samples_queue.get(block=True)
                try:
                    sample, test_count, success_count = item
                    self._write_sample(sample, test_count, success_count)
                except BaseException as exc:
                    log.debug("Processing sample failed: %s\n%s", str(exc), traceback.format_exc())
                    log.warning("Couldn't process sample, skipping")
                    continue

    def _write_sample(self, sample, test_count, success_count):
        line = json.dumps(sample.to_dict()) + "\n"
        self.out_stream.write(line.encode('utf-8'))
        self.out_stream.flush()

        report_pattern = "%s,Total:%d Passed:%d Failed:%d\n"
        failed_count = test_count - success_count
        sys.stdout.write(report_pattern % (sample.test_case, test_count, success_count, failed_count))
        sys.stdout.flush()


class JTLSampleWriter(LDJSONSampleWriter):
    def __init__(self, output_file):
        super(JTLSampleWriter, self).__init__(output_file)

    def __enter__(self):
        obj = super(JTLSampleWriter, self).__enter__()

        fieldnames = ["timeStamp", "elapsed", "Latency", "label", "responseCode", "responseMessage", "success",
                      "allThreads", "bytes"]
        endline = '\n'  # \r will be preprended automatically because out_stream is opened in text mode
        self.writer = csv.DictWriter(self.out_stream, fieldnames=fieldnames, dialect=csv.excel, lineterminator=endline,
                                     encoding='utf-8')
        self.writer.writeheader()
        self.out_stream.flush()

        return obj

    def _write_sample(self, sample, test_count, success_count):
        """
        :type sample: Sample
        :type test_count: int
        :type success_count: int
        """
        self._write_request_subsamples(sample)

    def _get_sample_type(self, sample):
        if sample.path:
            last = sample.path[-1]
            return last.type
        else:
            return None

    def _write_request_subsamples(self, sample):
        if self._get_sample_type(sample) == "request":
            self._write_single_sample(sample)
        elif sample.subsamples:
            for sub in sample.subsamples:
                self._write_request_subsamples(sub)
        else:
            self._write_single_sample(sample)

    def _write_single_sample(self, sample):
        """
        :type sample: Sample
        """
        bytes = sample.extras.get("responseHeadersSize", 0) + 2 + sample.extras.get("responseBodySize", 0)

        message = sample.error_msg
        if not message:
            message = sample.extras.get("responseMessage")
        if not message:
            for sample in sample.subsamples:
                if sample.error_msg:
                    message = sample.error_msg
                    break
                elif sample.extras.get("responseMessage"):
                    message = sample.extras.get("responseMessage")
                    break
        self.writer.writerow({
            "timeStamp": int(1000 * sample.start_time),
            "elapsed": int(1000 * sample.duration),
            "Latency": 0,  # TODO
            "label": sample.test_case,

            "bytes": bytes,

            "responseCode": sample.extras.get("responseCode"),
            "responseMessage": message,
            "allThreads": self.concurrency,  # TODO: there will be a problem aggregating concurrency for rare samples
            "success": "true" if sample.status == "PASSED" else "false",
        })
        self.out_stream.flush()


# noinspection PyPep8Naming
class ApiritifPlugin(Plugin):
    """
    Saves test results in a format suitable for Taurus.
    :type sample_writer: LDJSONSampleWriter
    """

    name = 'apiritif'
    enabled = False

    def __init__(self):
        super(ApiritifPlugin, self).__init__()
        self.controller = store.SampleController(log)
        apiritif.put_into_thread_store(controller=self.controller)  # parcel for smart_transactions
        self.stop_reason = ""

    def finalize(self, result):
        """
        After all tests
        """
        if not self.controller.test_count:
            raise RuntimeError("Nothing to test.")

    def beforeTest(self, test):
        """
        before test run
        """
        thread.clean_transaction_handlers()
        addr = test.address()  # file path, package.subpackage.module, class.method
        test_file, module_fqn, class_method = addr
        test_fqn = test.id()  # [package].module.class.method
        suite_name, case_name = test_fqn.split('.')[-2:]
        log.debug("Addr: %r", addr)
        log.debug("id: %r", test_fqn)

        if class_method is None:
            class_method = case_name

        description = test.shortDescription()
        self.controller.test_info = {
            "test_case": case_name,
            "suite_name": suite_name,
            "test_file": test_file,
            "test_fqn": test_fqn,
            "description": description,
            "module_fqn": module_fqn,
            "class_method": class_method}
        self.controller.beforeTest()  # create template of current_sample

    def startTest(self, test):
        self.controller.startTest()

    def stopTest(self, test):
        self.controller.stopTest()

    def afterTest(self, test):
        self.controller.afterTest()

    def addError(self, test, error):
        """
        when a test raises an uncaught exception
        :param test:
        :param error:
        :return:
        """
        # test_dict will be None if startTest wasn't called (i.e. exception in setUp/setUpClass)
        # status=BROKEN
        assertion_name = error[0].__name__
        error_msg = str(error[1]).split('\n')[0]
        error_trace = get_trace(error)
        if self.controller.current_sample is not None:
            self.controller.addError(assertion_name, error_msg, error_trace)
        else:  # error in test infrastructure (e.g. module setup())
            log.error("\n".join((assertion_name, error_msg, error_trace)))

    @staticmethod
    def isNormalShutdown(cls):
        cls_full_name = ".".join((cls.__module__, cls.__name__))
        ns_full_name = ".".join((NormalShutdown.__module__, NormalShutdown.__name__))
        return cls_full_name == ns_full_name

    def handleError(self, test, error):
        if self.isNormalShutdown(error[0]):
            self.add_stop_reason(error[1].args[0])  # remember it for run_nose() cycle
            return True
        else:
            return False

    def add_stop_reason(self, msg):
        if self.stop_reason:
            self.stop_reason += "\n"

        self.stop_reason += msg

    def addFailure(self, test, error):
        """
        when a test fails
        :param test:
        :param error:

        :return:
        """
        # status=FAILED
        self.controller.addFailure(error)

    def addSuccess(self, test):
        """
        when a test passes
        :param test:
        :return:
        """
        self.controller.addSuccess()


def cmdline_to_params():
    parser = OptionParser()
    parser.add_option('', '--concurrency', action='store', type="int", default=1)
    parser.add_option('', '--iterations', action='store', type="int", default=sys.maxsize)
    parser.add_option('', '--ramp-up', action='store', type="float", default=0)
    parser.add_option('', '--steps', action='store', type="int", default=sys.maxsize)
    parser.add_option('', '--hold-for', action='store', type="float", default=0)
    parser.add_option('', '--result-file-template', action='store', type="str", default="result-%s.csv")
    parser.add_option('', '--verbose', action='store_true', default=False)
    parser.add_option('', "--version", action='store_true', default=False)
    opts, args = parser.parse_args()
    log.debug("%s %s", opts, args)

    if opts.version:
        print(VERSION)
        sys.exit(0)

    params = Params()
    params.concurrency = opts.concurrency
    params.ramp_up = opts.ramp_up
    params.steps = opts.steps
    params.iterations = opts.iterations
    params.hold_for = opts.hold_for

    params.report = opts.result_file_template
    params.tests = args
    params.worker_count = 1  # min(params.concurrency, multiprocessing.cpu_count())
    params.verbose = opts.verbose

    return params


def setup_logging(params):
    logformat = "%(asctime)s:%(levelname)s:%(process)s:%(thread)s:%(name)s:%(message)s"
    apiritif.http.log.setLevel(logging.WARNING)
    if params.verbose:
        logging.basicConfig(level=logging.DEBUG, stream=sys.stdout, format=logformat)
    else:
        logging.basicConfig(level=logging.INFO, stream=sys.stdout, format=logformat)
    log.setLevel(logging.INFO)  # TODO: do we need to include apiritif debug logs in verbose mode?


def main():
    cmd_params = cmdline_to_params()
    setup_logging(cmd_params)
    supervisor = Supervisor(cmd_params)
    supervisor.start()
    supervisor.join()


if __name__ == '__main__':
    main()



# temporary exchanger for ApiritifPlugin stuff
import time
import traceback

import apiritif
from apiritif.samples import ApiritifSampleExtractor, Sample, PathComponent
from apiritif.utils import get_trace

writer = None


class SampleController(object):
    def __init__(self, log):
        self.current_sample = None  # todo: recreate it from plugin's template every transaction
        self.success_count = None
        self.log = log
        self.test_count = 0
        self.success_count = 0
        self.apiritif_extractor = ApiritifSampleExtractor()
        self.tran_mode = False  # it's regular test (without smart transaction) by default
        self.start_time = None
        self.end_time = None
        self.test_info = {}

    def beforeTest(self):
        self.current_sample = Sample(
            test_case=self.test_info["test_case"],
            test_suite=self.test_info["suite_name"],
            start_time=time.time(),
            status="SKIPPED")
        self.current_sample.extras.update({
            "file": self.test_info["test_file"],
            "full_name": self.test_info["test_fqn"],
            "description": self.test_info["description"]
        })
        module_fqn_parts = self.test_info["module_fqn"].split('.')
        for item in module_fqn_parts[:-1]:
            self.current_sample.path.append(PathComponent("package", item))
        self.current_sample.path.append(PathComponent("module", module_fqn_parts[-1]))

        if "." in self.test_info["class_method"]:  # TestClass.test_method
            class_name, method_name = self.test_info["class_method"].split('.')[:2]
            self.current_sample.path.extend([
                PathComponent("class", class_name),
                PathComponent("method", method_name)])
        else:  # test_func
            self.current_sample.path.append(PathComponent("func", self.test_info["class_method"]))

        self.log.debug("Test method path: %r", self.current_sample.path)

        self.test_count += 1

    def startTest(self):
        self.start_time = time.time()

    def stopTest(self, is_transaction=False):
        if self.tran_mode == is_transaction:
            self.end_time = time.time()

    def addError(self, assertion_name, error_msg, error_trace, is_transaction=False):
        if self.tran_mode == is_transaction:
            self.current_sample.add_assertion(assertion_name, {"args": [], "kwargs": {}})
            self.current_sample.set_assertion_failed(assertion_name, error_msg, error_trace)

    def addFailure(self, error, is_transaction=False):
        if self.tran_mode == is_transaction:
            assertion_name = error[0].__name__
            error_msg = str(error[1]).split('\n')[0]
            error_trace = get_trace(error)
            self.current_sample.add_assertion(assertion_name, {"args": [], "kwargs": {}})
            self.current_sample.set_assertion_failed(assertion_name, error_msg, error_trace)

    def addSuccess(self, is_transaction=False):
        if self.tran_mode == is_transaction:
            self.current_sample.status = "PASSED"
            self.success_count += 1

    def afterTest(self, is_transaction=False):
        if self.tran_mode == is_transaction:
            if self.end_time is None:
                self.end_time = time.time()
            self.current_sample.duration = self.end_time - self.current_sample.start_time

            samples_processed = self._process_apiritif_samples(self.current_sample)
            if not samples_processed:
                self._process_sample(self.current_sample)

            self.current_sample = None

    def _process_apiritif_samples(self, sample):
        samples = []

        # get list of events
        recording = apiritif.recorder.pop_events(from_ts=self.start_time, to_ts=self.end_time)

        try:
            if recording:
                # convert requests (events) to samples
                samples = self.apiritif_extractor.parse_recording(recording, sample)
        except BaseException as exc:
            self.log.debug("Couldn't parse recording: %s", traceback.format_exc())
            self.log.warning("Couldn't parse recording: %s", exc)

        for sample in samples:
            self._process_sample(sample)  # just write to disk

        return len(samples)

    def _process_sample(self, sample):
        writer.add(sample, self.test_count, self.success_count)



import os
import sys
from abc import ABCMeta, abstractmethod
from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules

from apiritif.utils import log

PLUGINS_PATH = 'PLUGINS_PATH'


def import_plugins():
    path = os.environ.get(PLUGINS_PATH, None)
    if not path:
        log.debug('Plugins PATH not found, continue without plugins')
        return

    # add plugins path to PYTHONPATH
    sys.path.append(path)

    package = Path(path).resolve().name
    log.info(f'Plugins package {package}')

    #  modules listing in the root package
    for (_, module_name, _) in iter_modules([path]):
        log.info(f'Importing module {module_name}')
        import_module(module_name)


class BaseActionHandler(metaclass=ABCMeta):

    YAML_ACTION_START = 'yaml_action_start'
    YAML_ACTION_END = 'yaml_action_end'
    TEST_CASE_START = 'test_case_start'
    TEST_CASE_STOP = 'test_case_stop'

    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def startup(self):
        pass

    @abstractmethod
    def handle(self, session_id, action_type, action):
        pass

    @abstractmethod
    def finalize(self):
        pass


class ActionHandlerFactory:
    registry = {}

    @classmethod
    def register(cls, name):
        def inner_wrapper(wrapped_class):
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create_handler(cls, name, **kwargs):
        if name not in cls.registry:
            log.warning('Handler %s does not exist in the registry', name)
            return

        exec_class = cls.registry[name]
        return exec_class(**kwargs)

    @classmethod
    def create_all(cls, **kwargs):
        return [cls.create_handler(name, **kwargs) for name in cls.registry]



"""
This is a toplevel package of Apiritif tool
Copyright 2017 BlazeMeter Inc.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .csv import CSVReaderPerThread
from .thread import put_into_thread_store, get_from_thread_store, external_handler, get_stage, set_stage
from .thread import get_transaction_handlers, set_transaction_handlers, get_iteration
from .http import http, transaction, transaction_logged, smart_transaction, recorder
from .http import Event, TransactionStarted, TransactionEnded, Request, Assertion, AssertionFailure
from .utilities import *
from .utils import headers_as_text, assert_regexp, assert_not_regexp, log



"""
This is a toplevel package of Apiritif tool

Copyright 2021 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
from OpenSSL import crypto
from datetime import datetime
from requests.adapters import HTTPAdapter
from urllib3.contrib.pyopenssl import PyOpenSSLContext

try:
    from ssl import PROTOCOL_TLS as ssl_protocol
except ImportError:
    from ssl import PROTOCOL_SSLv23 as ssl_protocol


class SSLAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        certificate_file_path = kwargs.pop('certificate_file_path', None)
        passphrase = kwargs.pop('passphrase', None)

        pkcs12_obj = CertificateReader.create_pkcs12_obj(certificate_file_path, passphrase)
        self.ssl_context = CertificateReader.create_ssl_context(pkcs12_obj)

        super(SSLAdapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        if self.ssl_context:
            kwargs['ssl_context'] = self.ssl_context
        return super(SSLAdapter, self).init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        if self.ssl_context:
            kwargs['ssl_context'] = self.ssl_context
        return super(SSLAdapter, self).proxy_manager_for(*args, **kwargs)


class CertificateReader:
    @staticmethod
    def create_pkcs12_obj(certificate_file_path, passphrase):
        """
        :param certificate_file_path: str
        :param passphrase: str
        :return: pkcs12_obj
        :rtype: OpenSSL.crypto.PKCS12
        """

        certificate_password = passphrase.encode('utf8')
        with open(certificate_file_path, 'rb') as pkcs12_file:
            certificate_data = pkcs12_file.read()

        if os.path.splitext(certificate_file_path)[-1].lower() == '.pem':
            pkcs12_obj = CertificateReader._create_openssl_cert_from_pem(certificate_data, certificate_password)
        else:
            pkcs12_obj = CertificateReader._create_openssl_cert_from_pkcs12(certificate_data, certificate_password)

        return pkcs12_obj

    @staticmethod
    def create_ssl_context(pkcs12_cert):
        """
        :param pkcs12_cert: OpenSSL.crypto.PKCS12
        :return: context
        :rtype: urllib3.contrib.pyopenssl.PyOpenSSLContext
        """

        cert = pkcs12_cert.get_certificate()
        CertificateReader._check_cert_not_expired(cert)

        context = PyOpenSSLContext(ssl_protocol)
        context._ctx.use_certificate(cert)

        ca_certs = pkcs12_cert.get_ca_certificates()
        if ca_certs:
            for ca_cert in ca_certs:
                CertificateReader._check_cert_not_expired(ca_cert)
                context._ctx.add_extra_chain_cert(ca_cert)

        context._ctx.use_privatekey(pkcs12_cert.get_privatekey())

        return context

    @staticmethod
    def _check_cert_not_expired(cert):
        cert_not_after = datetime.strptime(cert.get_notAfter().decode('ascii'), '%Y%m%d%H%M%SZ')
        if cert_not_after < datetime.utcnow():
            raise ValueError('SSL certificate expired')

    @staticmethod
    def _create_openssl_cert_from_pem(certificate_data, certificate_password):
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, certificate_data)
        key = crypto.load_privatekey(crypto.FILETYPE_PEM, certificate_data, passphrase=certificate_password)

        pkcs = crypto.PKCS12()
        pkcs.set_privatekey(key)
        pkcs.set_certificate(cert)
        return pkcs

    @staticmethod
    def _create_openssl_cert_from_pkcs12(certificate_data, certificate_password):
        return crypto.load_pkcs12(certificate_data, certificate_password)



"""
Data feeders for Apiritif.

Copyright 2018 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
import threading

import csv
from io import open
from itertools import cycle, islice
from chardet.universaldetector import UniversalDetector

import apiritif.thread as thread
from apiritif.utils import NormalShutdown

thread_data = threading.local()


class Reader(object):
    def read_vars(self):
        pass

    def get_vars(self):
        pass

    def close(self):
        pass


class CSVReaderPerThread(Reader):  # processes multi-thread specific
    def __init__(self, filename, fieldnames=None, delimiter=None, loop=True, quoted=None, encoding=None):
        self.filename = filename
        self.fieldnames = fieldnames
        self.delimiter = delimiter
        self.loop = loop
        self.quoted = quoted
        self.encoding = encoding

    def _get_csv_reader(self, create=False):
        csv_readers = getattr(thread_data, "csv_readers", None)
        if not csv_readers:
            thread_data.csv_readers = {}

        csv_reader = thread_data.csv_readers.get(id(self))
        if not csv_reader and create:
            csv_reader = CSVReader(
                filename=self.filename,
                fieldnames=self.fieldnames,
                step=thread.get_total(),
                first=thread.get_index(),
                delimiter=self.delimiter,
                loop=self.loop,
                quoted=self.quoted,
                encoding=self.encoding)

            thread_data.csv_readers[id(self)] = csv_reader

        return csv_reader

    def read_vars(self):
        self._get_csv_reader(create=True).read_vars()

    def close(self):
        csv_reader = self._get_csv_reader()
        if csv_reader:
            del thread_data.csv_readers[id(self)]
            csv_reader.close()

    def get_vars(self):
        csv_reader = self._get_csv_reader()
        if csv_reader:
            return csv_reader.get_vars()
        else:
            return {}


class CSVReader(Reader):
    def __init__(self, filename, step=1, first=0, fieldnames=None, delimiter=None, loop=True, quoted=None,
                 encoding=None):
        self.step = step
        self.first = first
        self.csv = {}
        format_params = {}
        if not encoding and quoted is None:
            with open(filename, 'rb') as bin_fds:
                if not encoding:
                    detector = UniversalDetector()
                    for line in bin_fds.readlines():
                        detector.feed(line)
                        if detector.done:
                            break
                    detector.close()
                    encoding = detector.result['encoding']
                    bin_fds.seek(0)

                if quoted is None:
                    header = bin_fds.readline()
                    header = header[:-1].decode(encoding=encoding)
                    match = re.match(r'.*["\']\w+["\'](.["\']\w+["\'])+', header)
                    quoted = True if match is not None else False
                    bin_fds.seek(0)
        format_params["quoting"] = csv.QUOTE_MINIMAL if quoted else csv.QUOTE_NONE

        self.fds = open(filename, 'r', encoding=encoding)

        if not delimiter:
            dialect = csv.Sniffer().sniff(self.fds.read())
            self.fds.seek(0)
            delimiter = dialect.delimiter
        format_params["delimiter"] = delimiter

        self._reader = csv.DictReader(self.fds, fieldnames=fieldnames, **format_params)
        if loop:
            self._reader = cycle(self._reader)

    def close(self):
        if self.fds is not None:
            self.fds.close()
        self._reader = None

    def read_vars(self):
        if not self._reader:
            return  # todo: exception?

        try:
            if not self.csv:  # first element
                self.csv = next(islice(self._reader, self.first, self.first + 1))
            else:  # next one
                self.csv = next(islice(self._reader, self.step - 1, self.step))
        except StopIteration:
            stop_reason = "Data source is exhausted: %s" % self.fds.name
            raise NormalShutdown(stop_reason)  # Just send it up

    def get_vars(self):
        return self.csv



"""
Public utility functions for Apiritif

Copyright 2017 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import base64
import random
import re
import string
import sys
from datetime import datetime
from uuid import uuid4

if sys.version_info < (3, 0):
    from urllib import quote_plus
else:
    from urllib.parse import quote_plus


def random_uniform(start, stop=None):
    return random.randrange(start, stop=stop)


def random_gauss(sigma, mu):
    return random.gauss(sigma, mu)


def random_string(size, chars=string.printable):
    return "".join(random.choice(chars) for _ in range(size))


class SimpleDateFormat(object):
    def __init__(self, format):
        self.format = format

    @staticmethod
    def _replacer(match):
        what = match.group(0)
        if what.startswith("y") or what.startswith("Y"):
            if len(what) < 4:
                return "%y"
            else:
                return "%Y"
        elif what.startswith("M"):
            return "%m"
        elif what.startswith("d"):
            return "%d"
        elif what.startswith("h"):
            return "%I"
        elif what.startswith("H"):
            return "%H"
        elif what.startswith("m"):
            return "%M"
        elif what.startswith("s"):
            return "%S"
        elif what.startswith("S"):
            return what
        elif what.startswith("E"):
            if len("E") <= 3:
                return "%a"
            else:
                return "%A"
        elif what.startswith("D"):
            return "%j"
        elif what.startswith("w"):
            return "%U"
        elif what.startswith("a"):
            return "%p"
        elif what.startswith("z"):
            return "%z"
        elif what.startswith("Z"):
            return "%Z"

    def format_datetime(self, datetime):
        letters = "yYMdhHmsSEDwazZ"  # TODO: moar
        regex = "(" + "|".join(letter + "+" for letter in letters) + ")"
        strftime_fmt = re.sub(regex, self._replacer, self.format)
        return datetime.strftime(strftime_fmt)


def format_date(format_string=None, datetime_obj=None):
    """
    Format a datetime object with Java SimpleDateFormat's-like string.

    If datetime_obj is not given - use current datetime.
    If format_string is not given - return number of millisecond since epoch.

    :param format_string:
    :param datetime_obj:
    :return:
    :rtype string
    """
    datetime_obj = datetime_obj or datetime.now()
    if format_string is None:
        seconds = int(datetime_obj.strftime("%s"))
        milliseconds = datetime_obj.microsecond // 1000
        return str(seconds * 1000 + milliseconds)
    else:
        formatter = SimpleDateFormat(format_string)
        return formatter.format_datetime(datetime_obj)


def base64_decode(encoded):
    decoded = base64.b64decode(encoded)
    return decoded.decode()


def base64_encode(plaintext):
    if not isinstance(plaintext, bytes):
        plaintext = plaintext.encode()
    encoded = base64.b64encode(plaintext)
    return encoded.decode()


def encode_url(chars):
    return quote_plus(chars)


def uuid():
    return str(uuid4())



"""

Copyright 2019 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from threading import local

from apiritif.action_plugins import BaseActionHandler

_total = 1
_thread_local = local()


def set_total(total):
    global _total
    _total = total


def get_total():
    global _total
    return _total


def set_index(index):
    _thread_local.index = index


def get_index():
    index = getattr(_thread_local, "index", 0)
    return index


def set_iteration(iteration):
    _thread_local.iteration = iteration


def get_iteration():
    iteration = getattr(_thread_local, "iteration", 0)
    return iteration


def get_stage():
    return get_from_thread_store("stage")


def set_stage(stage):
    put_into_thread_store(stage=stage)


def put_into_thread_store(*args, **kwargs):
    if args:
        _thread_local.args = args
    if kwargs:
        current_kwargs = getattr(_thread_local, "kwargs", {})
        current_kwargs.update(kwargs)
        _thread_local.kwargs = current_kwargs


def get_from_thread_store(names=None):
    if names and hasattr(_thread_local, "kwargs"):
        only_one = False
        if isinstance(names, str):
            names = [names]
            only_one = True
        kwargs = [_thread_local.kwargs.get(key) for key in names]
        if only_one:
            return kwargs[0]
        else:
            return kwargs

    elif hasattr(_thread_local, "args"):
        return _thread_local.args


def get_transaction_handlers():
    transaction_handlers = get_from_thread_store('transaction_handlers')
    return transaction_handlers


def clean_transaction_handlers():
    handlers = {'enter': [], 'exit': []}
    _thread_local.kwargs["transaction_handlers"] = handlers


def set_transaction_handlers(handlers):
    put_into_thread_store(transaction_handlers=handlers)


def external_handler(session_id, action_type, action):
    for handler in get_action_handlers():
        if isinstance(handler, BaseActionHandler):
            handler.handle(session_id, action_type, action)


def get_action_handlers():
    return get_from_thread_store("action_handlers")


# Deprecated
def external_log(log_line):
    pass


# Deprecated
def set_logging_handlers(handlers):
    pass


# Deprecated
def get_logging_handlers():
    pass


# Deprecated
def add_logging_handlers(methods=None):
    pass
